import aiofiles


async def read_text_file(filename: str) -> str:
    async with aiofiles.open(filename, mode='r') as file:
        content = await file.read()
    return content


async def write_text_file(filename: str, content: str):
    async with aiofiles.open(filename, mode='w') as file:
        await file.write(content)



from logging import Logger
from core.messagebus.messagebus import Publisher, Consumer
from core.rpc.rpc import Caller


def register_event(
    logger: Logger,
    consumer: Consumer,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Registering event handlers for "snake_module_name"')

    @consumer.register('hit_snake_module_name')
    async def hit_snake_module_name_event(message):
        logger.info(f'👋 Getting message: {message}')



from typing import Any
from logging import Logger
from core.messagebus.messagebus import Publisher
from core.rpc.rpc import Caller, Server


def register_rpc(
    logger: Logger,
    rpc_server: Server,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Registering RPC handlers for "snake_module_name"')

    @rpc_server.register('process_snake_module_name')
    async def process_snake_module_name(*args: Any, **kwargs: Any):
        return {
            'args': args, 'kwargs': kwargs
        }



from typing import Any, List, Mapping
from ._group import project_group
from ..task.decorator import python_task
from ..task.base_task import Group, BaseTask
from ..task_env.env import Env
from ..task_input.bool_input import BoolInput
from ..runner import runner
from ..helper.string.jinja import is_probably_jinja


@python_task(
    name='get-default-env',
    description='Get default values for project environments',
    inputs=[
        BoolInput(
            name='export',
            shortcut='e',
            description='Whether add export statement or not',
            default=True
        )
    ],
    group=project_group,
    runner=runner
)
async def get_default_env(*args: Any, **kwargs: Any) -> str:
    env_map: Mapping[str, str] = {}
    env_map = fetch_env_map_from_group(env_map, project_group)
    env_keys = list(env_map.keys())
    env_keys.sort()
    should_export = kwargs.get('export', True)
    export_prefix = 'export ' if should_export else ''
    return '\n'.join([
        f'{export_prefix}{key}={env_map[key]}' for key in env_keys
    ])


def fetch_env_map_from_group(
    env_map: Mapping[str, str], group: Group
) -> Mapping[str, str]:
    for task in group.tasks:
        env_map = fetch_env_map_from_task(env_map, task)
    for sub_group in group.children:
        sub_env_map: Mapping[str, str] = fetch_env_map_from_group(
            env_map, sub_group
        )
        env_map = cascade_env_map(env_map, sub_env_map)
    return env_map


def fetch_env_map_from_task(
    env_map: Mapping[str, str], task: BaseTask
):
    task_env_map: Mapping[str, str] = {}
    for env_file in task.env_files:
        envs = env_file.get_envs()
        task_env_map = add_envs_to_env_map(task_env_map, envs)
    task_env_map = add_envs_to_env_map(task_env_map, task.envs)
    env_map = cascade_env_map(env_map, task_env_map)
    for upstream in task.upstreams:
        task_env_map = fetch_env_map_from_task(env_map, upstream)
    for checker in task.checkers:
        task_env_map = fetch_env_map_from_task(env_map, checker)
    return env_map


def add_envs_to_env_map(
    env_map: Mapping[str, str], envs: List[Env]
) -> Mapping[str, str]:
    for env in envs:
        if (
            env.os_name == '' or
            env.default == '' or
            is_probably_jinja(env.default)
        ):
            if env.name.startswith('FASTAPP'):
                print(env)
            continue
        env_map[env.os_name] = env.default
    return env_map


def cascade_env_map(
    env_map: Mapping[str, str],
    other_env_map: Mapping[str, str]
) -> Mapping[str, str]:
    for key, value in other_env_map.items():
        if key in env_map:
            continue
        env_map[key] = value
    return env_map



from fastapi import FastAPI
from logging import Logger
from core.messagebus.messagebus import Publisher
from core.rpc.rpc import Caller


def register_api(
    logger: Logger,
    app: FastAPI,
    rpc_caller: Caller,
    publisher: Publisher
):
    logger.info('🥪 Registering API for "snake_module_name"')

    @app.get('/api/v1/kebab-module-name')
    async def get_snake_module_name():
        # Publish hit event
        await publisher.publish(
            'hit_snake_module_name', '/api/v1/kebab-module-name'
        )
        # Send RPC request
        result = await rpc_caller.call(
            'process_snake_module_name', 'hello', 'world', magic_number=42
        )
        return result



from typing import Any, List
from dotenv import dotenv_values
from ..._group import project_add_group
from ....task.task import Task
from ....task.decorator import python_task
from ....task.resource_maker import ResourceMaker
from ....runner import runner
from .._common import (
    default_module_inputs, project_dir_input, app_name_input,
    module_name_input, validate_project_dir, get_default_module_replacements,
    get_default_module_replacement_middlewares, new_app_scaffold_lock
)
from ....helper import util
from ....helper.codemod.add_import_module import add_import_module
from ....helper.codemod.add_function_call import add_function_call
from ....helper.file.text import read_text_file, write_text_file

import os
import jsons


current_dir = os.path.dirname(__file__)


@python_task(
    name='validate',
    inputs=[project_dir_input, app_name_input, module_name_input],
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get('project_dir')
    validate_project_dir(project_dir)
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    snake_module_name = util.to_snake_case(module_name)
    automation_dir = os.path.join(
        project_dir, '_automate', util.to_snake_case(app_name)
    )
    if not os.path.exists(automation_dir):
        raise Exception(
            f'Automation directory does not exist: {automation_dir}'
        )
    source_dir = os.path.join(
        project_dir, 'src', f'{util.to_kebab_case(app_name)}'
    )
    if not os.path.exists(source_dir):
        raise Exception(f'Source does not exist: {source_dir}')
    module_path = os.path.join(source_dir, 'module', snake_module_name)
    if os.path.exists(module_path):
        raise Exception(f'Module directory already exists: {module_path}')


copy_resource = ResourceMaker(
    name='copy-resource',
    inputs=default_module_inputs,
    upstreams=[validate],
    replacements=get_default_module_replacements(),
    replacement_middlewares=get_default_module_replacement_middlewares(),
    template_path=os.path.join(current_dir, 'template'),
    destination_path='{{ input.project_dir }}',
    scaffold_locks=[new_app_scaffold_lock],
    excludes=[
        '*/deployment/venv',
        '*/src/kebab-app-name/venv',
        '*/src/kebab-app-name/frontend/node_modules',
        '*/src/kebab-app-name/frontend/build',
    ]
)


@python_task(
    name='register-module',
    inputs=[project_dir_input, app_name_input, module_name_input],
    upstreams=[copy_resource]
)
async def register_module(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    kebab_app_name = util.to_kebab_case(app_name)
    snake_module_name = util.to_snake_case(module_name)
    import_module_path = '.'.join([
        'module', snake_module_name
    ])
    function_name = f'register_{snake_module_name}'
    main_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'main.py'
    )
    task.print_out(f'Read code from: {main_file_path}')
    code = await read_text_file(main_file_path)
    task.print_out(
        f'Add import "{function_name}" from "{import_module_path}" to the code'
    )
    code = add_import_module(
        code=code,
        module_path=import_module_path,
        resource=function_name
    )
    task.print_out(f'Add "{function_name}" call to the code')
    code = add_function_call(
        code=code,
        function_name=function_name,
        parameters=[]
    )
    task.print_out(f'Write modified code to: {main_file_path}')
    await write_text_file(main_file_path, code)


@python_task(
    name='create-app-config',
    inputs=[project_dir_input, app_name_input, module_name_input],
    upstreams=[copy_resource]
)
async def create_app_config(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    kebab_app_name = util.to_kebab_case(app_name)
    snake_module_name = util.to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    config_file_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'config.py'
    )
    config_code = '\n'.join([
        f'app_enable_{snake_module_name}_module = str_to_boolean(os.environ.get(',  # noqa
        f"    'APP_{upper_snake_module_name}', 'true'"
        '))'
    ])
    task.print_out(f'Read config from: {config_file_path}')
    code = await read_text_file(config_file_path)
    code += '\n' + config_code
    task.print_out(f'Write config to: {config_file_path}')
    await write_text_file(config_file_path, code)


@python_task(
    name='create-automation-config',
    inputs=[project_dir_input, app_name_input, module_name_input],
    upstreams=[copy_resource]
)
async def create_automation_config(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    snake_app_name = util.to_snake_case(app_name)
    snake_module_name = util.to_snake_case(module_name)
    json_modules_file_path = os.path.join(
        project_dir, '_automate', snake_app_name, 'config', 'modules.json'
    )
    task.print_out(f'Read json config from: {json_modules_file_path}')
    json_str = await read_text_file(json_modules_file_path)
    task.print_out(f'Add "{snake_module_name}" to json config')
    modules: List[str] = jsons.loads(json_str)
    modules.append(snake_module_name)
    json_str = jsons.dumps(modules)
    task.print_out(f'Write new json config to: {json_modules_file_path}')
    await write_text_file(json_modules_file_path, json_str)


@python_task(
    name='append-template-env',
    inputs=[project_dir_input, app_name_input, module_name_input],
    upstreams=[create_automation_config]
)
async def append_template_env(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    project_dir = kwargs.get('project_dir', '.')
    app_name = kwargs.get('app_name')
    module_name = kwargs.get('module_name')
    kebab_app_name = util.to_kebab_case(app_name)
    snake_app_name = util.to_snake_case(app_name)
    snake_module_name = util.to_snake_case(module_name)
    upper_snake_module_name = snake_module_name.upper()
    json_modules_file_path = os.path.join(
        project_dir, '_automate', snake_app_name, 'config', 'modules.json'
    )
    template_env_path = os.path.join(
        project_dir, 'src', kebab_app_name, 'src', 'template.env'
    )
    template_env_map = dotenv_values(template_env_path)
    app_port = int(template_env_map.get('APP_PORT', '8080'))
    task.print_out(f'Read json config from: {json_modules_file_path}')
    json_str = await read_text_file(json_modules_file_path)
    modules: List[str] = jsons.loads(json_str)
    module_app_port = app_port + len(modules)
    template_env_str = await read_text_file(template_env_path)
    template_env_str += '\n' + '\n'.join([
        f'APP_ENABLE_{upper_snake_module_name}_MODULE=true',
        f'APP_{upper_snake_module_name}_MODULE_PORT={module_app_port}',
    ])
    await write_text_file(template_env_path, template_env_str)


# TODO: add env to module_enabled.env <-- use the ones in src, because of docker-compose
# TODO: add env to module_disabled.env
# TODO: create service in docker compose
# TODO: create runner for local microservices


@python_task(
    name='fastapp-module',
    group=project_add_group,
    upstreams=[
        register_module,
        create_app_config,
        create_automation_config,
        append_template_env,
    ],
    runner=runner
)
async def add_fastapp_module(*args: Any, **kwargs: Any):
    task: Task = kwargs.get('_task')
    task.print_out('Success')



from fastapi.responses import HTMLResponse
import os

frontend_index_path = os.path.join('frontend', 'build', 'index.html')
with open(frontend_index_path, "r") as f:
    frontend_index_content = f.read()
frontend_index_response = HTMLResponse(
    content=frontend_index_content, status_code=200
)



from zrb import DockerComposeTask, runner
from zrb.builtin._group import project_group
from ._common import (
    RESOURCE_DIR, SKIP_CONTAINER_EXECUTION,
    app_container_checker, rabbitmq_checker, rabbitmq_management_checker,
    redpanda_console_checker, kafka_outside_checker,
    kafka_plaintext_checker, pandaproxy_outside_checker,
    pandaproxy_plaintext_checker,
    local_input, mode_input, host_input, https_input, image_input,
    compose_app_broker_type_env, compose_app_host_port_env,
    start_container_compose_profile_env, all_compose_profile_env, image_env,
    compose_rabbitmq_host_port_env, compose_rabbitmq_management_host_port_env,
    compose_redpanda_console_host_port_env,
    compose_kafka_outside_host_port_env,
    compose_pandaproxy_outside_host_port_env,
    compose_kafka_plaintext_host_port_env,
    compose_pandaproxy_plaintext_host_port_env
)
from .image import build_snake_app_name_image

compose_env_prefix = 'CONTAINER_ENV_PREFIX'
compose_envs = [
    compose_app_broker_type_env,
    compose_app_host_port_env,
    compose_rabbitmq_host_port_env,
    compose_rabbitmq_management_host_port_env,
    compose_redpanda_console_host_port_env,
    compose_kafka_outside_host_port_env,
    compose_kafka_plaintext_host_port_env,
    compose_pandaproxy_outside_host_port_env,
    compose_pandaproxy_plaintext_host_port_env
]

###############################################################################
# Task Definitions
###############################################################################

remove_snake_app_name_container = DockerComposeTask(
    icon='💨',
    name='remove-kebab-app-name-container',
    description='Rumove human readable app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    compose_cmd='down',
    compose_env_prefix=compose_env_prefix,
    envs=compose_envs + [
        image_env,
        all_compose_profile_env
    ],
)
runner.register(remove_snake_app_name_container)

start_snake_app_name_container = DockerComposeTask(
    icon='🐳',
    name='start-kebab-app-name-container',
    description='Start human readable app name container',
    group=project_group,
    inputs=[
        local_input,
        mode_input,
        host_input,
        https_input,
        image_input,
    ],
    skip_execution=SKIP_CONTAINER_EXECUTION,
    upstreams=[
        build_snake_app_name_image,
        remove_snake_app_name_container
    ],
    cwd=RESOURCE_DIR,
    compose_cmd='up',
    compose_env_prefix=compose_env_prefix,
    envs=compose_envs + [
        start_container_compose_profile_env,
        image_env,
    ],
    checkers=[
        app_container_checker,
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ]
)
runner.register(start_snake_app_name_container)

stop_snake_app_name_container = DockerComposeTask(
    icon='⛔',
    name='stop-kebab-app-name-container',
    description='Stop human readable app name container',
    group=project_group,
    cwd=RESOURCE_DIR,
    compose_cmd='stop',
    compose_env_prefix=compose_env_prefix,
    envs=compose_envs + [
        image_env,
        all_compose_profile_env
    ],
)
runner.register(stop_snake_app_name_container)



from fastapi import FastAPI, status
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from config import (
    app_enable_frontend, cors_allow_credentials, cors_allow_headers,
    cors_allow_methods, cors_allow_origin_regex, cors_allow_origins,
    cors_expose_headers, cors_max_age
)
from component.app_lifespan import app_state
from component.app_lifespan import app_lifespan
from component.frontend_index import frontend_index_response

app = FastAPI(lifespan=app_lifespan)

if app_enable_frontend:
    @app.middleware("http")
    async def catch_all(request, call_next):
        response = await call_next(request)
        if response.status_code == 404:
            return frontend_index_response
        return response


app.add_middleware(
    CORSMiddleware,
    allow_origins=cors_allow_origins,
    allow_origin_regex=cors_allow_origin_regex,
    allow_methods=cors_allow_methods,
    allow_headers=cors_allow_headers,
    allow_credentials=cors_allow_credentials,
    expose_headers=cors_expose_headers,
    max_age=cors_max_age,
)


@app.head('/liveness')
@app.get('/liveness')
def handle_liveness():
    if app_state.get_liveness():
        return JSONResponse(
            content={'message': 'Service is alive'},
            status_code=status.HTTP_200_OK
        )
    return JSONResponse(
        content={'message': 'Service is not alive'},
        status_code=status.HTTP_503_SERVICE_UNAVAILABLE
    )


@app.head('/readiness')
@app.get('/readiness')
def handle_readiness():
    if app_state.get_readiness():
        return JSONResponse(
            content={'message': 'Service is ready'},
            status_code=status.HTTP_200_OK
        )
    return JSONResponse(
        content={'message': 'Service is not ready'},
        status_code=status.HTTP_503_SERVICE_UNAVAILABLE
    )



from typing import List
from helper.conversion import str_to_boolean, str_to_logging_level
import os
import json

app_name = os.environ.get('APP_NAME', 'app')
app_logging_level = str_to_logging_level(
    os.environ.get('APP_LOGGING_LEVEL', 'INFO')
)
app_broker_type = os.environ.get('APP_BROKER_TYPE', 'mock')
app_host = os.environ.get('APP_HOST', '0.0.0.0')
app_port = int(os.environ.get('APP_PORT', '8080'))
app_reload = str_to_boolean(os.environ.get('APP_RELOAD', 'true'))

app_enable_rpc_server: bool = str_to_boolean(os.getenv(
    'APP_ENABLE_RPC_SERVER', 'true'
))

app_enable_message_consumer: bool = str_to_boolean(os.getenv(
    'APP_ENABLE_MESSAGE_CONSUMER', 'true'
))

app_enable_frontend: bool = str_to_boolean(os.getenv(
    'APP_ENABLE_FRONTEND', 'true'
))

app_enable_api: bool = str_to_boolean(os.getenv(
    'APP_ENABLE_API', 'true'
))

app_rmq_connection_string = os.environ.get(
    'APP_RMQ_CONNECTION', 'amqp://guest:guest@localhost/'
)

app_kafka_bootstrap_servers = os.environ.get(
    'APP_KAFKA_BOOTSTRAP_SERVERS', 'localhost:9092'
)

cors_allow_origins: List[str] = json.loads(os.getenv(
    'APP_CORS_ALLOW_ORIGINS', '["*"]'
))
cors_allow_origin_regex: str = os.getenv(
    'APP_CORS_ALLOW_ORIGIN_REGEX', ''
)
cors_allow_methods: List[str] = json.loads(os.getenv(
    'APP_CORS_ALLOW_METHODS', '["*"]'
))
cors_allow_headers: List[str] = json.loads(os.getenv(
    'APP_CORS_ALLOW_HEADERS', '["*"]'
))
cors_allow_credentials: bool = str_to_boolean(os.getenv(
    'APP_CORS_ALLOW_CREDENTIALS', 'false'
))
cors_expose_headers: bool = str_to_boolean(os.getenv(
    'APP_CORS_EXPOSE_HEADERS', 'false'
))
cors_max_age: int = int(os.getenv(
    'APP_CORS_MAX_AGE', '600'
))
